# -*- coding: utf-8 -*-
# Datei VTools.py

from visual import *
import visual.controls


#--------------------------- Allgemeine Tools ------------------------

TkosysObjs=[]  # Objekte des Koordinatensystems

def Tkosys(xmin=-10,xmax=10,ymin=-10,ymax=10,zmin=-10,zmax=10):
    """Tkosys shows three coordinate axis
    
    Default range is -10..10 for each axis, optionally xmin,xmax, ymin,ymax, zmin,zmax may be specified"""
    global TkosysObjs
    r=0.1
    xx=xmax-xmin
    yy=ymax-ymin
    zz=zmax-zmin
    TkosysObjs.append(cylinder(pos=(xmin,0,0),axis=(xx,0,0),radius=r,color=color.white))
    TkosysObjs.append(cylinder(pos=(0,ymin,0),axis=(0,yy,0),radius=r,color=color.white))
    TkosysObjs.append(cylinder(pos=(0,0,zmin),axis=(0,0,zz),radius=r,color=color.white))
    TkosysObjs.append(arrow(pos=(xmax-1,0,0),axis=(1,0,0),shaftwidth=r,color=color.white))
    TkosysObjs.append(arrow(pos=(0,ymin-1,0),axis=(0,1,0),shaftwidth=r,color=color.white))
    TkosysObjs.append(arrow(pos=(0,0,zmin-1),axis=(0,0,1),shaftwidth=r,color=color.white))
    TkosysObjs.append(label(pos=(xmax-1,0,0),text="x"))
    TkosysObjs.append(label(pos=(0,ymax-1,0),text="y"))
    TkosysObjs.append(label(pos=(0,0,zmax-1),text="z"))
    for x in range(-10,11): TkosysObjs.append(sphere(pos=(x,0,0),radius=2*r,color=color.blue))
    for y in range(-10,11): TkosysObjs.append(sphere(pos=(0,y,0),radius=2*r,color=color.blue))
    for z in range(-10,11): TkosysObjs.append(sphere(pos=(0,0,z),radius=2*r,color=color.blue))


def halfSphere(pos=(0,0,0),radius=1,color=color.white):
    """Construct a half sphere
    
    Arguments are position of center and radius"""
    obj=convex(color=color)
    L = []
    for t in arange(0,pi,0.2):
        for s in arange(0,pi,0.1):
            L.append((radius*cos(t)*sin(s), radius*sin(t)*sin(s), radius*cos(s)))
    obj.pos = L
    return obj


def knee(startpos=vector(0,0,0),axis=vector(1,0,0),r=1.0,angle=90):
    return helix(pos=startpos, axis=axis, radius=r,thickness=2*r,coils=angle/360.0)

# GUI-Tools

Tcontrols=None  # Window for buttons etc.

def TshowButtonPanel():
    """TshowButtonPanel()
    
    Shows a panels for buttons and sliders (see addButton, addSlider)"""
    global Tcontrols
    Tcontrols=visual.controls.controls(width=120, height=400, range=200, title='VTools')
    Tcontrols.nextpos=0
    Tcontrols.defaultWidth=100
    Tcontrols.defaultHeight=35

class TButton(visual.controls.button):
    def __init__(self,pos=(0,0),width=100, height=40,text="..", cmd=None):
        visual.controls.button.__init__(self,pos=pos,width=width,height=height,text=text,action=cmd)
    def setCommand(self,cmd): self.action=cmd

def addButton(text='...',cmd=None):
    """addButton(text='...',cmd=None)
    
    addButtons adds a button to the control window. The button is labeld with text.
    The functiom cmd will be called (without arguments) when the user clicks
    the button. The command can also be set with setCommand."""
    global Tcontrols
    if Tcontrols==None: return None
    Tcontrols.nextpos=Tcontrols.nextpos+Tcontrols.defaultHeight
    return TButton(pos=(0,Tcontrols.nextpos),
                   width=Tcontrols.defaultWidth, 
                   height=Tcontrols.defaultHeight,
                   text=text, cmd=cmd)

class TSlider(visual.controls.slider):
    def __init__(self,pos=(0,0),width=100, height=40, minimum=0, maximum=maximum, cmd=None):
        visual.controls.slider.__init__(self,pos=pos,min=minimum, max=maximum,
                                        width=width,height=height)
        if not(cmd==None): self.action=cmd
        self.axis=vector(1,0,0)
        self.length=int(Tcontrols.defaultWidth*0.5)
    def setCommand(self,cmd): self.action=cmd


def addSlider(minimum=0, maximum=100, cmd=None):
    """addSlider - adds a slider to the Visual Controls Window
    
    addSlider(minimum=0, maximum=100, cmd=None)
    mnimum and maximum specify the range of values
    The attribute value gives the current value
    The method setCommand defines the function (without parameters) to called 
    upon change of the sliders value by the user."""
    global Tcontrols
    if Tcontrols==None: return None
    Tcontrols.nextpos=Tcontrols.nextpos+Tcontrols.defaultHeight
    return TSlider(pos=(0,Tcontrols.nextpos), #-Tcontrols.defaultWidth/2
                   width=Tcontrols.defaultWidth/10, 
                   height=Tcontrols.defaultHeight/2,
                   minimum=minimum, maximum=maximum, cmd=cmd)


#---------------------------- 3D Turtle -------------------------------

def turtleform(size=1):
    """turtleform(size) - build a turtle of given size (default 1)
    
    """
    parts=[]
    parts.append(halfSphere((0,0,0),size,color=color.green))
    parts.append(cylinder(pos=(0.6*size,0,0.6*size), radius=0.15*size, axis=(0,-0.3*size,0),color=(174/255.0,145/255.0,0)))
    parts.append(cylinder(pos=(0.6*size,0,-0.6*size), radius=0.15*size,axis=(0,-0.3*size,0),color=(174/255.0,145/255.0,0)))
    parts.append(cylinder(pos=(-0.6*size,0,0.6*size), radius=0.15*size,axis=(0,-0.3*size,0),color=(174/255.0,145/255.0,0)))
    parts.append(cylinder(pos=(-0.6*size,0,-0.6*size),radius=0.15*size, axis=(0,-0.3*size,0),color=(174/255.0,145/255.0,0)))
    parts.append(sphere(pos=(0.85*size,0.2*size, 0.2*size),radius=0.2*size,color=color.blue))
    parts.append(sphere(pos=(0.85*size,0.2*size, -0.2*size),radius=0.2*size,color=color.blue))
    parts.append(cone(pos=(0.9*size,0.05*size, 0),radius=0.3*size,axis=(size*0.6,0,0),color=color.red))
    return parts

def wormform(size=1):
    """wormform(size) - build a 'turtle' looking like a worm 
    
    """
    parts=[]
    parts.append(sphere(pos=(0,0,0),radius=size/1.5,color=color.green))
    parts.append(sphere(pos=(-size/1.5,0,0),radius=size/2.0,color=color.green))
    parts.append(sphere(pos=(-size,0,0),radius=size/3.0,color=color.green))
    parts.append(sphere(pos=(0.75*size,0.2*size, 0.2*size),radius=0.2*size,color=color.blue))
    parts.append(sphere(pos=(0.75*size,0.2*size, -0.2*size),radius=0.2*size,color=color.blue))
    parts.append(cone(pos=(0.8*size,0.05*size, 0),radius=0.3*size,axis=(size*0.6,0,0),color=color.red))
    return parts



class Turtle3D(frame):
    """Turtle3D - a turtle in space
    
    ..."""
    def __init__(self,size=1.0,model=turtleform):
        """Turtle3D constructs a 3D turtle
        
        Optional parameters: 
            model a function of one parameter (size) that builds the turtle
            size turtle's size (default 1.0)"""
        
        frame.__init__(self)
        self.dir=vector(1,0,0)   # forward direcions - local coordinates 
        self.lrdir=vector(0,0,1) # left-right-direction 
        self.dudir=vector(0,1,0) # down-up-direction
        self.turtleFrame = frame()
        self.color=color.green   # Farbe der Spur
        self.thickness=0.1       # Dicke der Spur
        self.drawnObjects=[]
        self.penDownState=True
        self.animated=True
        self.animationSteps=50 # Zahl der Zwischenschitte
        self.animationTime=0.8 # Dauer der Animation
        # Zusammenstellung der Turtle
        self.parts=model(size)
        for part in self.parts: part.frame=self

    def setVisible(self,yesno):
        """setVisible(yesno) sets the trutle to be visible or not
        
        yesno is a boolean value"""
        assert(type(yesno)==bool)
        for obj in self.parts: obj.visible=yesno
    def setAnimated(self,yesno): 
        """setAnimated(yesno) sets the trutle to move slowly (animated)
        
        yesno is a boolean value
        The attribute animationTime gives the total time (in seconds) of the animation
        and animationSteps defines how many frames the animation has"""
        assert(type(yesno)==bool)
        self.animated=yesno   
    def penUp(self):
        """penUp() sets the pen up so that no trail is drawn"""
        self.penDownState=False 
    def penDown(self): 
        """penDown() sets the pen down so that a trail is drawn"""
        self.penDownState=True
    def setColor(self,col):
        """setColor
        
        Sets the color in color conventions of visual i.e. color.red or (0.5,1,0.8) """
        self.color=col
    def setThickness(self,th): 
        """setThickness(number)
        
        Sets the thickness of the turtle's trail (default 0.1)""" 
        assert(type(th)==int or type(th)==float)
        self.thickness=th
    def goto(self,pos): 
        """goto a specified position
        
        Example: turtle.goto((4,2,1.5))  """
        if type(pos)==tuple: pos=vector(pos)
        assert(type(pos)==vector)
        if self.animated:
            self.animated=False
            oldpos=vector(self.pos)
            dist=(pos-self.pos)/float(self.animationSteps)
            for i in range(self.animationSteps):
                rate(self.animationSteps/float(self.animationTime))
                self.goto(oldpos+i*dist)
            self.animated=True
        else:
            oldpos=vector(self.pos)
            self.pos=pos
            ax=pos-oldpos
            if self.penDownState:
                self.drawnObjects.append(cylinder(pos=oldpos,axis=ax,radius=self.thickness,color=self.color))
    def lookdir(self,dir):
        """lookdir sets the direction the turtle is looking to
                
        Example: turtle.lookdir((1,2,0))"""
        if type(dir)==tuple: dir=vector(dir)
        assert(type(dir)==vector)
        drehachse=cross(self.dir,dir)
        self.turn(degrees(acos(dot(dir/mag(dir),self.dir/mag(self.dir)))),drehachse)
    def backward(self,dist):
        """backward(dist) moves the turtle backward
        
        Example: turtle.backward(2.5)"""
        self.forward(-dist)
    def forward(self,dist):
        """forward(dist) moves the turtle forward
        
        Example: turtle.forward(2.5)"""
        assert(type(dist)==int or type(dist)==float)
        if self.animated:
            self.animated=False
            step=dist/float(self.animationSteps)
            for i in range(self.animationSteps):
                rate(self.animationSteps/float(self.animationTime))
                self.forward(step)
            self.animated=True
        else:
            oldpos=vector(self.pos)
            ax=dist*self.dir/mag(self.dir)
            newpos=oldpos+ax
            self.pos=newpos
            if self.penDownState:
                self.drawnObjects.append(cylinder(pos=oldpos,axis=ax,radius=self.thickness,color=self.color))
    def turnDown(self,angle): 
        """turnDown(angle) turns the turtle down
        
        The angle is given in degrees
        Example: turtle.turnDown(45)"""
        self.turnUp(-angle)
    def turnUp(self,angle):
        """turnUp(angle) turns the turtle up
        
        The angle is given in degrees
        Example: turtle.turnUp(45)"""
        self.rotate(angle=radians(angle), axis=self.lrdir, origin=self.pos)
        self.dir = rotate(self.dir, angle=radians(angle), axis=self.lrdir)
        self.dudir = rotate(self.dudir, angle=radians(angle), axis=self.lrdir)
    def turnRight(self,angle):
        """turnRight(angle) turns the angle right 
        
        The angle is given in degrees.
        Example: turtle.turnRight(45)"""
        self.turnLeft(-angle)
    def turnLeft(self,angle):
        """turnLeft(angle) turns the angle left 
        
        The angle is given in degrees.
        Example: turtle.turnLeft(45)"""
        assert(type(angle)==int or type(angle)==float)
        self.rotate(angle=radians(angle), axis=self.dudir, origin=self.pos)
        self.dir = rotate(self.dir, angle=radians(angle), axis=self.dudir)
        self.lrdir = rotate(self.lrdir, angle=radians(angle), axis=self.dudir)
    def turn(self,angle,axis):
        """turn(angle,axis)
        
        The turtle is rotated along the given axis by the given angle"""
        if type(axis)==tuple: axis=vector(axis)
        assert(type(axis)==vector)        
        self.rotate(angle=radians(angle), axis=axis, origin=self.pos)
        self.dir = rotate(self.dir, angle=radians(angle), axis=axis)
        self.lrdir = rotate(self.lrdir, angle=radians(angle), axis=axis)
        self.dudir = rotate(self.dudir, angle=radians(angle), axis=axis)
        
        
#---------------- Interaktivitaet ----------------------------------


VclickEvent=None      # ggf Funktion mit zwei Parametern (Maus-Pos, None oder getrof. Obj) 
VpressEvent=None
VdragEvent=None
VreleaseEvent=None

VLoopProc=None

VclickObjEvents={} 
VpressObjEvents={} 
VdragObjEvents={}
VreleaseObjEvents={}

VkeyEvents={}

def setClickEvent(cmd):
    """setClickEvent(cmd) sets the command that is executed when the user clicks the mouse
    
    
    cmd gets two parameters: The position of the click and the object hit or None"""
    global VclickEvent
    VclickEvent=cmd
def setPressEvent(cmd):
    """setPressEvent(cmd) sets the command that is executed when the user presses the mouse
    
    
    cmd gets two parameters: The position of the click and the object hit or None"""
    global VpressEvent
    VpressEvent=cmd
def setDragEvent(cmd):
    """setDragEvent(cmd) sets the command that is executed when the user drags the mouse
    
    
    cmd gets two parameters: The position of the click and the object hit or None"""
    global VdragEvent
    VdragEvent=cmd
def setReleaseEvent(cmd):
    """setReleaseEvent(cmd) sets the command that is executed when the user releases the mouse
    
    
    cmd gets two parameters: The position of the click and the object hit or None"""
    global VreleaseEvent
    VreleaseEvent=cmd

def setObjectsClickEvent(obj,cmd): # cmd ist Funktion von einem Parameter (pos) 
    """setObjectsClickEvent(obj,cmd) sets the command to be executed when obj is clicked
    
    cmd must be a function that takes one parameter (the position of the mouse)"""
    global VclickObjEvents
    VclickObjEvents[obj]=cmd
def setObjectsDragEvent(obj,cmd): # cmd ist Funktion von einem Parameter (pos) 
    """setObjectsDragEvent(obj,cmd) sets the command to be executed when obj is dragged
    
    cmd must be a function that takes one parameter (the position of the mouse)"""
    global VdragObjEvents
    VdragObjEvents[obj]=cmd
def setObjectsPressEvent(obj,cmd): # cmd ist Funktion von einem Parameter (pos) 
    """setObjectsPressEvent(obj,cmd) sets the command to be executed when the mouse is pressed on obj 
    
    cmd must be a function that takes one parameter (the position of the mouse)"""
    global VpressObjEvents
    VpressObjEvents[obj]=cmd
def setObjectsReleaseEvent(obj,cmd): # cmd ist Funktion von einem Parameter (pos) 
    """setObjectsReleaseEvent(obj,cmd) sets the command to be executed when the mouse on obj is released
    
    cmd must be a function that takes one parameter (the position of the mouse)"""
    global VreleaseObjEvents
    VreleaseObjEvents[obj]=cmd
def setKeyEvent(k,cmd): # cmd ist Funktion mit zwei Parametern (Maus-Pos, None oder getrof. Obj)  
    """setKeyEvent(k,cmd) sets the command to be executed when the uses presses key k
    
    cmd must be a function that takes two parameters (the position of the mouse and the object at the mouse)
    Exampes:
    def linksProzedur(mousePos,object): turtle.pos+=(1,2,3)
    setKeyEvent("left",linksProzedur)
    setKeyEvent("a",linksProzedur)"""
    global VkeyEvents
    VkeyEvents[k]=cmd

def setLoopProc(cmd):
    """setLoopProc(cmd) sets the loop procedure to be cmd
    
    This sets a proceudre that is called without parameters in each iteration of the main loop."""
    global VLoopProc
    VLoopProc=cmd
    
Vrate=25 # FPS 

def setVLoopRate(r):
    """setVLoopRate(r) sets the loop rate to r
    
    This means that the main event loop function will be executed at most r times per second"""
    global Vrate
    Vrate=r

def Vmainloop():
    """Vmainloop - this starts the main event loop
    
    This function only ends when the visual grapbics window is closed"""
    run=1
    while run:
        if Tcontrols!=None: Tcontrols.interact()
        if scene.mouse.events:
            ev=scene.mouse.getevent()
            if ev.click: 
                if not(VclickEvent==None):
                    VclickEvent(scene.mouse.pos,scene.mouse.pick)
                if not(scene.mouse.pick==None):
                    obj=scene.mouse.pick
                    if VclickObjEvents.has_key(obj):
                        VclickObjEvents[obj](scene.mouse.pos)
            if ev.press: 
                if not(VpressEvent==None):
                    VpressEvent(scene.mouse.pos,scene.mouse.pick)
                if not(scene.mouse.pick==None):
                    obj=scene.mouse.pick
                    if VpressObjEvents.has_key(obj):
                        VpressObjEvents[obj](scene.mouse.pos)
            if ev.drag:
                if not(VdragEvent==None):
                    VdragEvent(scene.mouse.pos,scene.mouse.pick)
                if not(scene.mouse.pick==None):
                    obj=scene.mouse.pick
                    if VdragObjEvents.has_key(obj):
                        VdragObjEvents[obj](scene.mouse.pos)
            if ev.release: 
                if not(VreleaseEvent==None):
                    VreleaseEvent(scene.mouse.pos,scene.mouse.pick)
                if not(scene.mouse.pick==None):
                    obj=scene.mouse.pick
                    if VreleaseObjEvents.has_key(obj):
                        VreleaseObjEvents[obj](scene.mouse.pos)
        if scene.kb.keys: # event waiting to be processed?
            s = scene.kb.getkey() # get keyboard info
            if VkeyEvents.has_key(s):
                VkeyEvents[s](scene.mouse.pos,scene.mouse.pick)
        if not(VLoopProc==None):
            VLoopProc(scene)
        rate(Vrate)

import thread
def VmainloopThread():
    """VmainloopThread - tstarts the main event loop in a new thread
    
    """
    thread.start_new_thread(Vmainloop,())


print "VTools Version 1.0 loaded."
